/*
 *  TOPPERS Software
 *      Toyohashi Open Platform for Embedded Real-Time Systems
 *
 *  Copyright (C) 2006-2016 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2018,2019 by Naoki Saito
 *             Nagoya Municipal Industrial Research Institute, JAPAN
 *
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  $Id: bcm2837.h 386 2019-04-17 03:13:00Z nmir-saito $
 */
#ifndef TOPPERS_BCM2837_H
#define TOPPERS_BCM2837_H

#include <sil.h>

/*
 *  割込み番号（intno）の定義
 */
// Core interrupts
#define TINTNO_CNTPS        0    // CNTPSIRQ
#define TINTNO_CNTPNS       1    // CNTPNSIRQ
#define TINTNO_CNTHP        2    // CNTHPIRQ
#define TINTNO_CNTV         3    // CNTVIRQ
#define TINTNO_MBOX0        4    // Mailbox 0
#define TINTNO_MBOX1        5    // Mailbox 1
#define TINTNO_MBOX2        6    // Mailbox 2
#define TINTNO_MBOX3        7    // Mailbox 3
#define TINTNO_PMU          9    // PMU
#define TINTNO_AXI         10    // AXI-outstanding
#define TINTNO_LTIMER      11    // Local timer

// Basic interrupts
#define TINTNO_ARMTIMER    32    // ARM timer
#define TINTNO_MBOX        33    // ARM mailbox
#define TINTNO_ARMDRBELL0  34    // ARM doorbell 0
#define TINTNO_ARMDRBELL1  35    // ARM doorbell 1
#define TINTNO_GPU0HALTED  36    // GPU0 halted
#define TINTNO_GPU1HALTED  37    // GPU1 halted
#define TINTNO_ILACSS0     38    // Illegal access type 1
#define TINTNO_ILACSS1     39    // Illegal access type 0

// GPU0 interrupts
#define TINTNO_AUX         93    // Aux

// GPU1 interrupts
#define TINTNO_I2CSPISLV  107
#define TINTNO_PWA0       109
#define TINTNO_PWA1       110
#define TINTNO_SMI        112
#define TINTNO_GPIO0      113
#define TINTNO_GPIO1      114
#define TINTNO_GPIO2      115
#define TINTNO_GPIO3      116
#define TINTNO_I2C        117
#define TINTNO_SPI        118
#define TINTNO_PCM        119
#define TINTNO_UART       121


/*
 *  割込み番号の最小値と最大値
 */
#define TMIN_INTNO              UINT_C(0)
#define TMAX_INTNO              UINT_C(127)

/*
 *  メモリマップドI/Oのベースアドレス
 */
#define MMIO_BASE               0x3F000000

/*
 *  割込み関連レジスタ
 */
#define IRQ_PEND_B              0x3f00b200
#define IRQ_PEND_1              0x3f00b204
#define IRQ_PEND_2              0x3f00b208
#define ENABLE_IRQ_1            0x3f00b210
#define ENABLE_IRQ_2            0x3f00b214
#define ENABLE_IRQ_B            0x3f00b218
#define DISABLE_IRQ_1           0x3f00b21C
#define DISABLE_IRQ_2           0x3f00b220
#define DISABLE_IRQ_B           0x3f00b224

#define CORE0_IRQ_SRC           0x40000060


/*
 * GPIO
 */
#define GPFSEL0         (MMIO_BASE+0x00200000)
#define GPFSEL1         (MMIO_BASE+0x00200004)
#define GPFSEL2         (MMIO_BASE+0x00200008)
#define GPFSEL3         (MMIO_BASE+0x0020000C)
#define GPFSEL4         (MMIO_BASE+0x00200010)
#define GPFSEL5         (MMIO_BASE+0x00200014)
#define GPSET0          (MMIO_BASE+0x0020001C)
#define GPSET1          (MMIO_BASE+0x00200020)
#define GPCLR0          (MMIO_BASE+0x00200028)
#define GPLEV0          (MMIO_BASE+0x00200034)
#define GPLEV1          (MMIO_BASE+0x00200038)
#define GPEDS0          (MMIO_BASE+0x00200040)
#define GPEDS1          (MMIO_BASE+0x00200044)
#define GPHEN0          (MMIO_BASE+0x00200064)
#define GPHEN1          (MMIO_BASE+0x00200068)
#define GPPUD           (MMIO_BASE+0x00200094)
#define GPPUDCLK0       (MMIO_BASE+0x00200098)
#define GPPUDCLK1       (MMIO_BASE+0x0020009C)


/*
 *  UARTレジスタのベースアドレス
 */
#define RPI_UART0_BASE  (MMIO_BASE + 0x00201000)

/*
 *  ボーレート設定
 */
#define RPI_UART_IBRD_38400     0x27U    // 38400bps
#define RPI_UART_FBRD_38400     0x04U
#define RPI_UART_IBRD_115200    0x2U    // 115200bps
#define RPI_UART_FBRD_115200    0xbU

/*
 * メールボックス関連
 *    https://github.com/raspberrypi/firmware/wiki/Mailbox-property-interface
 */
#define VIDEOCORE_MBOX  (MMIO_BASE+0x0000B880)

#define MBOX_READ       (VIDEOCORE_MBOX+0x0)
#define MBOX_POLL       (VIDEOCORE_MBOX+0x10)
#define MBOX_SENDER     (VIDEOCORE_MBOX+0x14)
#define MBOX_STATUS     (VIDEOCORE_MBOX+0x18)
#define MBOX_CONFIG     (VIDEOCORE_MBOX+0x1C)
#define MBOX_WRITE      (VIDEOCORE_MBOX+0x20)
#define MBOX_RESPONSE   0x80000000
#define MBOX_FULL       0x80000000
#define MBOX_EMPTY      0x40000000

#define MBOX_REQUEST    0

/* チャネル */
#define MBOX_CH_POWER   0
#define MBOX_CH_FB      1
#define MBOX_CH_VUART   2
#define MBOX_CH_VCHIQ   3
#define MBOX_CH_LEDS    4
#define MBOX_CH_BTNS    5
#define MBOX_CH_TOUCH   6
#define MBOX_CH_COUNT   7
#define MBOX_CH_PROP    8

/* tags */
#define MBOX_TAG_GETSERIAL      0x10004
#define MBOX_TAG_LAST           0

#define MBOX_REQUEST    0

/* channels */
#define MBOX_CH_POWER   0
#define MBOX_CH_FB      1
#define MBOX_CH_VUART   2
#define MBOX_CH_VCHIQ   3
#define MBOX_CH_LEDS    4
#define MBOX_CH_BTNS    5
#define MBOX_CH_TOUCH   6
#define MBOX_CH_COUNT   7
#define MBOX_CH_PROP    8

/* tags */
#define MBOX_TAG_GETSERIAL      0x10004
#define MBOX_TAG_SETCLKRATE     0x38002
#define MBOX_TAG_LAST           0

#ifndef TOPPERS_MACRO_ONLY

/*
 * mailbox 呼び出し
 *    注意：第2引数のアドレスは16バイトアライメントであること
 */
Inline int_t mbox_call(unsigned char ch, volatile uint32_t *mbox)
{
    uint32_t r = (((uint32_t)((uint64_t)mbox) & ~0xF) | (ch & 0xF));

    // 使えるようになるまで待つ
    do{
        Asm("nop");
    } while(sil_rew_mem((uint32_t *)MBOX_STATUS) & MBOX_FULL);

    // チャネルの識別子を付加してメッセージのアドレスを書き込み
    sil_wrw_mem((uint32_t *)MBOX_WRITE , r);

    while(1) {
        // 応答があるまで待つ
        do{
            Asm("nop");
        } while(sil_rew_mem((uint32_t *)MBOX_STATUS) & MBOX_EMPTY);

        if(r == sil_rew_mem((uint32_t *)MBOX_READ))
            return mbox[1] == MBOX_RESPONSE;
    }
    return 0;
}

#endif /* TOPPERS_MACRO_ONLY */
#endif /* TOPPERS_BCM2837_H */
